import sys
from threading import Thread
from time import sleep
from typing import Iterable

from PySide6.QtCore import (
    QObject,
    QCoreApplication,
    QUrl,
    qInstallMessageHandler,
    Slot,
    Signal,
)
from PySide6.QtGui import Qt, QIcon
from PySide6.QtQml import QQmlApplicationEngine
from PySide6.QtWidgets import QApplication

import dvpn.res  # noqa
from dvpn.config.constants import PublicVars, CLI_RESOLVE
from dvpn.logger import qt_message_handler
from dvpn.modules.tools import connect
from dvpn.vpns.base import VpnCli


class Bridge(QObject):
    periodic_thread = None
    connectedVPNs: Iterable[str] = set()
    changingVPNs = set()

    disconnectBtnChangeEnabled = Signal(bool, str, name="disconnectBtnChangeEnabled")
    connectStatusChange = Signal(str, bool, bool, name="connectStatusChange")
    disconnectChange = Signal(str, bool, bool, name="disconnectChange")

    @Slot(result=list)
    def get_available_cli(self) -> list:
        return list(CLI_RESOLVE.keys())

    def periodic_check(self):
        if self.periodic_thread is not None:
            return False
        t = Thread(target=self._periodic_check, name="Periodic status Check", daemon=True)
        self.periodic_thread = t
        t.start()
        return True

    def _periodic_check(self):
        while True:
            for vpn_name in list(PublicVars().credentials.keys()):

                # Wait for previous operation to finish
                while vpn_name in self.changingVPNs:
                    sleep(0.1)

                creds = PublicVars().credentials[vpn_name]
                cli_type = CLI_RESOLVE[creds["selectedVpn"]]
                cli_instance = cli_type(creds["cliPath"])
                state = cli_instance.get_state(vpn_name).lower()

                if "disconnected" in state:
                    self.disconnect_notify(vpn_name, cli_instance, False)
                    break

                elif "connected" in state and vpn_name not in self.connectedVPNs:
                    self.disconnectBtnChangeEnabled.emit(True, vpn_name)
                    break

                # Limit CPU usage
                sleep(0.1)

            sleep(0.5)

    @Slot(str)
    def log(self, text):
        print(text)

    @Slot(result="QVariantMap")
    def list_vpn(self) -> dict:
        return PublicVars().credentials

    @Slot(str, result="QVariantMap")
    def get_vpn_details(self, vpn_name: str) -> dict:
        if vpn_name not in PublicVars().credentials.keys():
            return {}
        details = PublicVars().credentials[vpn_name]
        # details["cliPath"] = QUrl.fromLocalFile(details["cliPath"]).toString()
        return details

    @Slot("QVariantMap", result=bool)
    def add_vpn(self, obj: dict) -> bool:
        vpn_name = obj["VPN Name"]
        if vpn_name not in PublicVars().credentials.keys():
            PublicVars()[vpn_name] = obj
            return True
        return False

    @Slot(str, result=str)
    def get_vpn_default_cli(self, vpn_name):
        cli: VpnCli = CLI_RESOLVE[vpn_name]
        return cli.get_default_cli_path()

    @Slot(str, result=list)
    def get_vpn_fields(self, vpn_name):
        cli = CLI_RESOLVE[vpn_name]
        return cli.fields

    @Slot(str)
    def connect(self, vpn_name: str):
        # noinspection PyUnresolvedReferences
        self.changingVPNs.add(vpn_name)
        self.connectStatusChange.emit(vpn_name, False, True)

        vpn_conf = PublicVars()[vpn_name]
        cli = CLI_RESOLVE[vpn_conf["selectedVpn"]](vpn_conf["cliPath"])
        t = Thread(
            target=lambda: connect(cli, vpn_conf["VPN Name"], self),
            name=f"Connecting {vpn_conf['VPN Name']}",
            daemon=True,
        )
        t.start()

    @Slot(str, "QVariantMap")
    def edit(self, vpn_name: str, contents: dict):
        tmp = PublicVars().credentials
        tmp[vpn_name] = contents
        PublicVars().credentials = tmp

    @Slot(str)
    def delete(self, vpn_name: str):
        tmp = PublicVars().credentials
        tmp.pop(vpn_name, None)
        PublicVars().credentials = tmp

    @Slot(str)
    def connected_notify(self, vpn_name: str):
        self.connectedVPNs.add(vpn_name)

    @Slot(str)
    def disconnect(self, vpn_name: str):
        # noinspection PyUnresolvedReferences
        self.disconnectChange.emit(vpn_name, True, True)
        vpn_conf = PublicVars()[vpn_name]
        cli = CLI_RESOLVE[vpn_conf["selectedVpn"]](vpn_conf["cliPath"])
        t = Thread(
            target=lambda: self.disconnect_notify(vpn_name, cli),
            name=f"Disconnecting {vpn_name}",
            daemon=True,
        )
        t.start()

    @Slot()
    def reset(self):
        for host in set(self.connectedVPNs):
            self.changingVPNs.add(host)
            creds = PublicVars().credentials[host]
            cli_type = CLI_RESOLVE[creds["selectedVpn"]]
            cli = cli_type(str(creds["cliPath"]))
            t = Thread(
                target=lambda: self.disconnect_notify(host, cli),
                name=f"Disconnecting {host}",
                daemon=True,
            )
            t.start()

    def disconnect_notify(self, vpn_name, cli, reset=True):
        self.disconnectChange.emit(vpn_name, False, True)
        if reset:
            cli.reset(host=vpn_name)
        # noinspection PyUnresolvedReferences
        self.disconnectChange.emit(vpn_name, False, False)
        self.connectedVPNs.discard(vpn_name)
        self.changingVPNs.discard(vpn_name)


def main():
    qInstallMessageHandler(qt_message_handler)
    QCoreApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    QCoreApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)
    app = QApplication(sys.argv)
    app.setOrganizationName("Jiri Otoupal")
    app.setOrganizationDomain("https://github.com/jiri-otoupal/dievpn")
    app.setWindowIcon(QIcon(":/icons/dievpn.ico"))
    app.setApplicationName("DieVpn")
    engine = QQmlApplicationEngine()
    bridge = Bridge()
    # bridge.periodic_check()
    context = engine.rootContext()
    context.setContextProperty("con", bridge)
    engine.load(QUrl("qrc:/qml/main.qml"))
    if not engine.rootObjects():
        sys.exit(-1)
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



from time import sleep
from typing import Optional

from dvpn.config.constants import PublicVars
from dvpn.vpns.base import VpnCli


def connect(cli: VpnCli, host: str, bridge: Optional["Bridge"]) -> (bool, dict):
    while len(bridge.connectedVPNs):
        sleep(0.1)
    creds = PublicVars().credentials[host]
    vpncli = type(cli)(str(creds["cliPath"]))
    try:
        out = vpncli.connect(creds)
        if bridge:
            bridge.connectStatusChange.emit(host, out[0], False)
            bridge.connectedVPNs.add(host)
    except Exception as ex:
        print(
            "DieVpn encountered problem with anyconnect, can be cause by stuck "
            "ovpn agent from previous instance or already running cli try to check"
            f" for other cli or anyconnect processes or reboot computer Exception {ex}"
        )
        return False, {"exception": str(ex)}
    finally:
        bridge.changingVPNs.discard(host)



import logging
import subprocess
import sys
from pathlib import Path
from time import sleep

from dvpn.vpns.base import VpnCli

if sys.platform == "win32":
    import wexpect
else:
    import pexpect as wexpect


class ViscosityCLI(VpnCli):
    fields = [
        {"name": "VPN Name", "placeholderText": "Name of VPN in Viscosity"},
    ]
    cli_path_win = Path("C:\\") / "Program Files" / "Viscosity" / "ViscosityCC.exe"

    cli_path_osx = Path("/opt/cisco/anyconnect/bin/vpn")

    def get_state(self, name: str) -> str:
        return subprocess.check_output(
            [
                self.cli_path if self.cli_path else self.get_default_cli_path(),
                "getstate",
                name,
            ]
        ).decode()

    def get_connected_vpn(self):
        pass

    def reset(self, cli_path=None, host: str = None):
        print("...Disconnecting")

        pipe = wexpect.spawn(
            command=cli_path if cli_path else self.get_default_cli_path(),
            args=["disconnect", "all" if host is None else host],
            encoding="utf-8",
        )
        output = pipe.readline()

        while host is not None and "Disconnected" not in self.get_state(host):
            sleep(0.1)

    def __connect(self, host) -> dict:
        print(f">> Connecting to {host}")
        self.process_pipe = wexpect.spawn(
            command=self.cli_path if self.cli_path else self.get_default_cli_path(),
            args=["connect", f"{host}"],
            encoding="utf-8",
            timeout=15,
        )
        output = self.process_pipe.readline()

        while (
            "Connected" not in (state_conn := self.get_state(host))
            and "Disconnected" not in state_conn
        ):
            sleep(0.1)

        logging.info("".join(output))
        connected = "Connected" in self.get_state(host)
        return {"connected": connected, "reason": "VPN Error", "log": output}

    def connect(self, creds: dict) -> (bool, dict):
        print("~ Resetting connection")
        self.reset()

        try:
            stat = self.__connect(creds["VPN Name"])
        except wexpect.TIMEOUT as ex:
            stat = {"reason": "invalid credentials"}

        if stat.get("connected", False):
            print("<Successfully Connected>")
            return True, stat
        else:
            print("Connection failed :(")
            print(f"Reason {''.join(stat.get('log', ['Unable to read stdout']))}")
            return False, stat



class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        cls_instances = cls._instances.get(cls) or []
        matching_instances = list(
            filter(
                lambda x: x["args"] == args and x["kwargs"] == kwargs,
                cls_instances,
            )
        )
        if len(matching_instances) == 1:
            return matching_instances[0]["instance"]
        else:
            instance = super().__call__(*args, **kwargs)
            cls_instances.append({"instance": instance, "args": args, "kwargs": kwargs})
            cls._instances[cls] = cls_instances
            return instance



from pathlib import Path

RESOURCE_DIRS = [
    Path(d) for d in ["icons", "qml", "qml/custom", "images", "qml/modals", "js"]
]

with open("main.qrc", "w", encoding="utf-8") as f:

    def fprint(string=""):
        print(string, file=f)

    fprint('<!DOCTYPE RCC><RCC version="1.0">')
    fprint("  <qresource>")

    # main files
    fprint(f"    <!-- Main Interface files -->")
    for res_dir in RESOURCE_DIRS:
        fprint(f"    <!-- {res_dir}/ -->")
        for file in res_dir.iterdir():
            if not file.is_file():
                continue
            fprint(f"    <file>{file}</file>")
        fprint()

    fprint("  </qresource>")
    fprint("</RCC>")



import logging
import os
import re
import subprocess
import sys
from pathlib import Path

from dvpn.vpns.base import VpnCli

if sys.platform == "win32":
    import wexpect
else:
    import pexpect as wexpect


class AnyConnectCLI(VpnCli):
    fields = [
        {"name": "VPN Name", "placeholderText": "My Zoo VPN"},
        {"name": "Host", "placeholderText": "favourite-zoo.com"},
        {"name": "Username", "placeholderText": "giraffe"},
        {"name": "Password", "placeholderText": "******", "sensitive": True},
    ]
    cli_path_win = (
            Path("C:\\")
            / "Program Files (x86)\\"
            / "Cisco\\"
            / "Cisco AnyConnect Secure Mobility Client\\vpncli.exe"
    )

    cli_path_osx = Path("/opt/cisco/anyconnect/bin/vpn")

    def get_state(self, name: str) -> str:
        stdout = subprocess.check_output([self.cli_path, "state", name]).decode()
        # Match and remove dot on end
        vpn_state = re.findall("(.*state:*.) (.*)", stdout)[0][1]

        return vpn_state.replace("\r", "")

    def get_connected_vpn(self):
        stdout = subprocess.check_output([self.cli_path, "stats"]).decode()
        # Match and remove dot on end
        vpn_name_index = re.findall(".*onnected to*. .*", stdout)[0][:-1]

        return vpn_name_index.split(" ")[1]

    def reset(self, cli_path=None, host: str = None):
        print("...Disconnecting")
        arguments = ["disconnect"]

        if host:
            arguments.append(host)

        pipe = wexpect.spawn(
            command=cli_path if cli_path else self.get_default_cli_path(),
            args=arguments,
            encoding="utf-8",
        )
        output = pipe.readlines()

        if "disconnected" in "".join(output[-3:-1]).lower():
            print("Disconnected")
        else:
            print("Not Connected")

    def __connect(self, host, username, password, **kwargs) -> dict:
        print(f">> Connecting to {host}")
        self.process_pipe = wexpect.spawn(
            command=self.cli_path,
            args=["connect", f"{host}"],
            encoding="utf-8",
            timeout=15,
        )

        print("     ... Waiting for VPN to complete its chores")
        self.process_pipe.expect(".*sername*.")
        print("     > Entering Username")
        self.process_pipe.sendline(username)
        self.process_pipe.expect(".*assword*.")
        print("     > Entering Password")
        self.process_pipe.sendline(password)

        if kwargs.get("banner", False):
            self.process_pipe.expect(".*accept*.")
            self.process_pipe.sendline("y")

        output = self.process_pipe.readlines()

        logging.info("".join(output))
        connected = "connected" in "".join(output[-3:-1]).lower()
        return {"connected": connected, "reason": "VPN Error", "log": output}

    def connect(self, creds: dict) -> (bool, dict):
        print("~ Resetting connection")
        self.reset()

        try:
            stat = self.__connect(
                host=creds["Host"],
                username=creds["Username"],
                password=creds["Password"],
                **creds,
            )
        except wexpect.TIMEOUT as ex:
            stat = {"reason": "invalid credentials"}

        if stat.get("connected", False):
            print("<Successfully Connected>")
            return True, stat
        else:
            print("Connection failed :(")
            print(f"Reason {''.join(stat.get('log', ['Unable to read stdout']))}")
            return False, stat



import json
import logging
from builtins import dict
from typing import Optional

from dvpn.config.paths import secret_path
from dvpn.vpns.anyconnect import AnyConnectCLI
from dvpn.vpns.tunnelblick import TunnelblickCLI
from dvpn.vpns.viscosity import ViscosityCLI

CONNECTED_CLI = None

DEFAULT_TITLE = "Die VPN Control"

CLI_RESOLVE = {"AnyConnect": AnyConnectCLI, "Viscosity": ViscosityCLI,
               "Tunnelblick": TunnelblickCLI}


class PublicVars:
    _credentials = {}
    instance = None

    def __init__(self):
        self._credentials = {}
        if secret_path.exists():
            self.load_vars()

    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = object.__new__(cls)
            return cls.instance
        else:
            return cls.instance

    @property
    def credentials(self) -> dict:
        if not secret_path.exists():
            self.save_vars({})
        else:
            self.load_vars()
        return self._credentials

    def save_vars(self, variables: Optional[dict] = None):
        if variables is None:
            vars_save = PublicVars().credentials
        else:
            vars_save = variables

        with open(str(secret_path), "w") as fp:
            json.dump(vars_save, fp, indent=4)

    def load_vars(self):
        with open(str(secret_path), "r") as fp:
            self._credentials = json.load(fp)
            for key, item in self._credentials.items():
                if "selectedVpn" not in item.keys():
                    logging.warning("Applying patch for older version")
                    tmp = dict()
                    tmp["VPN Name"] = key
                    tmp["selectedVpn"] = "AnyConnect"
                    tmp["cliPath"] = self._credentials[key].pop("cli_path")
                    tmp["Host"] = self._credentials[key].pop("host")
                    tmp["Username"] = self._credentials[key].pop("username")
                    tmp["Password"] = self._credentials[key].pop("password")
                    tmp["banner"] = self._credentials[key].pop("banner")
                    self._credentials[key] = tmp

        self.save_vars(self._credentials)

    @credentials.setter
    def credentials(self, updated: dict):
        self.save_vars(updated)

    def __getitem__(self, item):
        return self._credentials[item]

    def __setitem__(self, key, value):
        self._credentials[key] = value
        self.save_vars(self._credentials)



import logging

from PySide6 import QtCore


def config_logger(level=logging.INFO):
    import logging
    from rich.logging import RichHandler

    FORMAT = "%(message)s"
    logging.basicConfig(
        level=level, handlers=[RichHandler()], datefmt="[%X]", format=FORMAT
    )


def qt_message_handler(mode, context, message):
    try:
        if mode == QtCore.QtInfoMsg:
            logging.info(
                "{} at: {} line: {}".format(message, context.file, context.line)
            )
        elif mode == QtCore.QtWarningMsg:
            logging.warning(
                "{} at: {} line: {}".format(message, context.file, context.line)
            )
        elif mode == QtCore.QtCriticalMsg:
            logging.critical(
                "{} at: {} line: {}".format(message, context.file, context.line)
            )
        elif mode == QtCore.QtFatalMsg:
            logging.fatal(
                "{} at: {} line: {}".format(message, context.file, context.line)
            )
        else:
            logging.debug(
                "{} at: {} line: {}".format(message, context.file, context.line)
            )
    except Exception as ex:
        logging.fatal(ex.__str__())



import abc
import os

from dvpn.modules.vpn_cli_singleton import SingletonMeta


class VpnCli(metaclass=SingletonMeta):
    def __init__(self, cli_path):
        print("\nWelcome to DieVPN\n")
        print(
            "Make sure to kill all VPN clients before usage, as cli would collide with it"
        )
        print(
            "If this is stuck, make sure you are NOT running it in virtual environment !"
        )
        self.excluded_fields = []
        self.process_pipe = None
        self.cli_path = cli_path

    @property
    @abc.abstractmethod
    def cli_path_win(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def cli_path_osx(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_state(self, name: str) -> str:
        """
        Gets state of VPN
        :param name:
        :return: Disconnected | Connected | ...
        """
        pass

    @classmethod
    def get_default_cli_path(cls):
        if os.name == "nt":
            return str(cls.cli_path_win)
        return str(cls.cli_path_osx)

    @abc.abstractmethod
    def get_connected_vpn(self):
        pass

    @classmethod
    @abc.abstractmethod
    def reset(cls, cli_path=None, host: str = None):
        pass

    @abc.abstractmethod
    def __connect(self, **kwargs) -> dict:
        pass

    @abc.abstractmethod
    def connect(self, creds: dict) -> (bool, dict):
        pass
